package limax.node.js.modules;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.net.StandardSocketOptions;
import java.nio.ByteBuffer;
import java.nio.channels.AsynchronousServerSocketChannel;
import java.nio.channels.AsynchronousSocketChannel;
import java.nio.channels.CompletionHandler;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;

import javax.script.Invocable;

import limax.node.js.Buffer;
import limax.node.js.EventLoop;
import limax.node.js.EventLoop.Callback;
import limax.node.js.EventLoop.EventObject;
import limax.node.js.Module;

public final class Net implements Module {
	private final EventLoop eventLoop;
	private final Invocable invocable;

	private static InetSocketAddress createSocketAddress(String host, int port, boolean server) {
		if (host != null)
			return new InetSocketAddress(host, port);
		if (server)
			new InetSocketAddress(port);
		return new InetSocketAddress("localhost", port);
	}

	public Net(EventLoop eventLoop) {
		this.eventLoop = eventLoop;
		this.invocable = eventLoop.getInvocable();
	}

	private static class ServerManager {
		private final static Map<SocketAddress, ServerManager> managers = new ConcurrentHashMap<>();
		private final List<Server> servers = new ArrayList<>();
		private int round = 0;

		private final AsynchronousServerSocketChannel assc;

		private synchronized boolean completed(AsynchronousSocketChannel asc) {
			int count = servers.size();
			if (count == 0)
				return false;
			for (int i = 0; i < count; i++) {
				if (servers.get(round = (round + 1) % count).completed(asc))
					return true;
			}
			try {
				asc.close();
			} catch (Exception e) {
			}
			return true;
		}

		private synchronized boolean failed(Throwable exc) {
			int count = servers.size();
			if (count == 0)
				return false;
			servers.get(round = (round + 1) % servers.size()).failed(exc);
			return true;
		}

		private synchronized void addServer(Server server) {
			servers.add(server);
		}

		private synchronized boolean removeServer(Server server) {
			servers.remove(server);
			if (!servers.isEmpty())
				return false;
			try {
				assc.close();
			} catch (Exception e) {
			}
			return true;
		}

		private ServerManager(SocketAddress sa, int backlog) throws Exception {
			assc = AsynchronousServerSocketChannel.open();
			assc.setOption(StandardSocketOptions.SO_REUSEADDR, true);
			assc.bind(sa, backlog);
			accept();
		}

		private void accept() {
			try {
				assc.accept(null, new CompletionHandler<AsynchronousSocketChannel, Object>() {

					@Override
					public void completed(AsynchronousSocketChannel asc, Object attachment) {
						if (ServerManager.this.completed(asc))
							accept();
					}

					@Override
					public void failed(Throwable exc, Object attachment) {
						if (ServerManager.this.failed(exc))
							accept();
					}
				});
			} catch (Exception e) {
			}
		}

		static synchronized void listen(Server server, int backlog) throws Exception {
			ServerManager sm = managers.get(server.sa);
			if (sm == null)
				managers.put(server.sa, sm = new ServerManager(server.sa, backlog));
			sm.addServer(server);
		}

		static synchronized void close(Server server) {
			ServerManager sm = managers.get(server.sa);
			if (sm != null && sm.removeServer(server))
				managers.remove(server.sa);
		}
	}

	public class Server {
		private final Set<AsynchronousSocketChannel> channels = Collections.newSetFromMap(new ConcurrentHashMap<>());
		private final Object THIS;
		private int maxConnections = Integer.MAX_VALUE;
		private boolean closed = false;
		private Callback cb;
		private EventObject evo;
		private InetSocketAddress sa;

		Server(Object THIS) throws Exception {
			this.THIS = THIS;
		}

		public Object[] address() {
			try {
				InetAddress addr = sa.getAddress();
				return new Object[] { sa.getPort(), addr instanceof Inet4Address ? "IPv4" : "IPv6",
						addr.getHostAddress() };
			} catch (Exception e) {
				return new Object[] {};
			}
		}

		boolean completed(AsynchronousSocketChannel asc) {
			if (channels.size() >= maxConnections)
				return false;
			channels.add(asc);
			cb.call(null, asc);
			return true;
		}

		void failed(Throwable exc) {
			cb.call(exc);
		}

		public void listen(int port, String host, int backlog, Object callback) throws Exception {
			if (evo != null)
				cb.call(new Exception("Duplicate listen is forbidden"));
			this.evo = eventLoop.createEventObject();
			this.cb = eventLoop.createCallback(callback);
			this.sa = createSocketAddress(host, port, true);
			ServerManager.listen(this, backlog);
			invocable.invokeMethod(THIS, "emit", "listening");
		}

		private void gracefulClose() {
			if (channels.isEmpty()) {
				try {
					invocable.invokeMethod(THIS, "emit", "close");
				} catch (Exception e) {
				} finally {
					evo.queue();
				}
			}
		}

		public void close() {
			ServerManager.close(this);
			closed = true;
			gracefulClose();
		}

		void detach(AsynchronousSocketChannel sc) {
			channels.remove(sc);
			if (closed)
				gracefulClose();
		}

		public void getConnections(Object callback) {
			eventLoop.execute(callback, r -> r.add(channels.size()));
		}

		public boolean isListening() {
			return evo != null;
		}

		public int getMaxConnections() {
			return maxConnections;
		}

		public void setMaxConnections(int maxConnections) {
			this.maxConnections = maxConnections;
		}

		public void ref() {
			evo.ref();
		}

		public void unref() {
			evo.unref();
		}
	}

	public class Socket {
		private final AsynchronousSocketChannel asc;
		private final Queue<ByteBuffer> wqueue = new ArrayDeque<>();
		private final Queue<Callback> wcallback = new ArrayDeque<>();
		private boolean writting = false;
		private final EventObject evo;
		private long timeout = Integer.MAX_VALUE;

		Socket(AsynchronousSocketChannel asc) {
			this(asc, eventLoop.createEventObject());
		}

		Socket(AsynchronousSocketChannel asc, EventObject evo) {
			this.asc = asc;
			this.evo = evo;
		}

		public Object[] address() {
			try {
				InetSocketAddress sa = (InetSocketAddress) asc.getLocalAddress();
				InetAddress addr = sa.getAddress();
				return new Object[] { sa.getPort(), addr instanceof Inet4Address ? "IPv4" : "IPv6",
						addr.getHostAddress() };
			} catch (Exception e) {
				return new Object[] {};
			}
		}

		public void destroy(Server owner) {
			try {
				asc.shutdownOutput(); // send FIN avoid RST on Windows
				asc.close();
				if (owner != null)
					owner.detach(asc);
			} catch (Exception e) {
			} finally {
				evo.queue();
			}
		}

		public String getLocalAddress() {
			try {
				return ((InetSocketAddress) asc.getLocalAddress()).getAddress().getHostAddress();
			} catch (Exception e) {
				return null;
			}
		}

		public Integer getLocalPort() {
			try {
				return ((InetSocketAddress) asc.getLocalAddress()).getPort();
			} catch (Exception e) {
				return null;
			}
		}

		public String getRemoteAddress() {
			try {
				return ((InetSocketAddress) asc.getRemoteAddress()).getAddress().getHostAddress();
			} catch (Exception e) {
				return null;
			}
		}

		public String getRemoteFamily() {
			try {
				return ((InetSocketAddress) asc.getRemoteAddress()).getAddress() instanceof Inet4Address ? "IPv4"
						: "IPv6";
			} catch (IOException e) {
				return null;
			}
		}

		public Integer getRemotePort() {
			try {
				return ((InetSocketAddress) asc.getRemoteAddress()).getPort();
			} catch (Exception e) {
				return null;
			}
		}

		public void setKeepAlive(boolean on) {
			try {
				asc.setOption(StandardSocketOptions.SO_KEEPALIVE, on);
			} catch (Exception e) {
			}
		}

		public void setNoDelay(boolean on) {
			try {
				asc.setOption(StandardSocketOptions.TCP_NODELAY, on);
			} catch (Exception e) {
			}
		}

		public void setTimeout(int timeout) {
			this.timeout = timeout;
		}

		public void read(Integer size, Object callback) {
			ByteBuffer bb = ByteBuffer.allocateDirect(size == null ? 16384 : size);
			Callback cb = eventLoop.createCallback(callback);
			try {
				asc.read(bb, timeout, TimeUnit.MILLISECONDS, null, new CompletionHandler<Integer, Object>() {
					@Override
					public void completed(Integer result, Object attachment) {
						bb.flip();
						cb.call(null, result, new Buffer(bb));
					}

					@Override
					public void failed(Throwable exc, Object attachment) {
						cb.call(exc);
					}
				});
			} catch (Exception e) {
				cb.call(e);
			}
		}

		private synchronized void writedone() {
			writting = false;
			while (!wqueue.isEmpty() && !wqueue.peek().hasRemaining()) {
				wqueue.poll();
				Callback cb = wcallback.poll();
				if (cb != null)
					cb.call(new Object[] {});
			}
			if (!wqueue.isEmpty())
				writeloop();
		}

		private synchronized void writefail(Throwable exc) {
			writting = false;
			wcallback.stream().filter(Objects::nonNull).forEach(cb -> cb.call(exc));
			wqueue.clear();
			wcallback.clear();
		}

		private synchronized void writeloop() {
			if (writting)
				return;
			writting = true;
			ByteBuffer[] srcs = wqueue.toArray(new ByteBuffer[0]);
			try {
				asc.write(srcs, 0, srcs.length, timeout, TimeUnit.MILLISECONDS, null,
						new CompletionHandler<Long, Object>() {

							@Override
							public void completed(Long result, Object attachment) {
								writedone();
							}

							@Override
							public void failed(Throwable exc, Object attachment) {
								writefail(exc);
							}
						});
			} catch (Exception e) {
				writefail(e);
			}
		}

		public synchronized void write(Buffer buffer, Object callback) {
			wqueue.add(buffer.toByteBuffer());
			wcallback.add(eventLoop.createCallback(callback));
			writeloop();
		}

		public synchronized void writev(Object[] srcs, Object callback) {
			int len = srcs.length;
			Callback cb = eventLoop.createCallback(callback);
			if (len == 0) {
				cb.call(null, 0);
				return;
			}
			for (int i = 0; i < len; i++)
				wqueue.add(((Buffer) srcs[i]).toByteBuffer());
			while (--len > 0)
				wcallback.add(null);
			wcallback.add(cb);
			writeloop();
		}

		public void shutdownInput() {
			try {
				asc.shutdownInput();
			} catch (Exception e) {
			}
		}

		public void shutdownOutput() {
			try {
				asc.shutdownOutput();
			} catch (Exception e) {
			}
		}

		public void ref() {
			evo.ref();
		}

		public void unref() {
			evo.unref();
		}
	}

	public Socket createSocket(AsynchronousSocketChannel sc) {
		return new Socket(sc);
	}

	public void connect(int port, String host, String localAddress, Integer localPort, Object callback) {
		AsynchronousSocketChannel asc;
		Callback cb = eventLoop.createCallback(callback);
		try {
			asc = AsynchronousSocketChannel.open();
			if (localAddress != null || localPort != null)
				asc.bind(createSocketAddress(localAddress, localPort == null ? 0 : localPort, true));
			EventObject evo = eventLoop.createEventObject();
			asc.connect(createSocketAddress(host, port, false), null, new CompletionHandler<Void, Object>() {

				@Override
				public void completed(Void result, Object attachment) {
					cb.call(null, new Socket(asc, evo));
				}

				@Override
				public void failed(Throwable exc, Object attachment) {
					cb.call(exc);
					evo.queue();
				}
			});
		} catch (Exception e) {
			cb.call(e);
		}
	}

	public Server createServer(Object THIS) throws Exception {
		return new Server(THIS);
	}
}
